from contextvars import copy_context
from ophis.globals import app_context, Request, Response


def test_context():
    app_context.inject('foo', 'bar', "MY_SCOPE")
    app_context.inject('some_value', 'This is a global', 'MY_SCOPE')
    app_context.inject('rqu', 'value', 'REQUEST')
    assert app_context.resolve('MY_SCOPE') == {
        'foo': 'bar',
        'some_value': 'This is a global'
    }

    assert list(app_context.scopes()) == ['GLOBAL', 'MY_SCOPE', 'REQUEST']
    assert list(app_context.keys_in_scope('MY_SCOPE')) == ['foo', 'some_value']


def test_request():
    request = Request()
    request.headers = {'content-type': 'application/json'}
    request.cookies = ['abc=123']
    ctx = copy_context()

    def change_context():
        request.headers = {'content-type': 'text/plain'}
        request.cookies = ['efg=456']
        return {
            'headers': request.headers,
            'cookies': request.cookies
        }

    assert ctx.run(change_context) == {
        'headers': {'content-type': 'text/plain'},
        'cookies': ['efg=456']
    }
    assert request.headers == {'content-type': 'application/json'}
    assert request.cookies == ['abc=123']

    try:
        request.farts
        assert False
    except AttributeError:
        assert True


def test_response():
    response = Response()
    response.headers = {'content-type': 'text/plain'}
    response.status_code = 404
    response.body = "Not found"
    ctx = copy_context()

    def change_context():
        response.headers = {'content-type': 'application/json'}
        response.status_code = 200
        response.body = {'message': 'Hello World!'}
        return {
            'headers': response.headers,
            'statusCode': response.status_code,
            'body': response.body
        }

    assert ctx.run(change_context) == {
        'headers': {'content-type': 'application/json'},
        'statusCode': 200,
        'body': {'message': 'Hello World!'}
    }

    assert response.status_code == 404
    assert response.headers == {'content-type': 'text/plain'}
    assert response.body == "Not found"



from contextvars import ContextVar
import logging
import os


logger = logging.getLogger(__name__)


class _ContextContainer:
    """
    Base class for context aware values. Extensions of this class
    define fields that fundamentally ContextVar with tags and values.
    """
    def __init__(self, name, fields) -> None:
        self.__context = {}
        for tag in fields:
            self.__context[tag] = ContextVar(f'{name}_{tag}')

    def __getattr__(self, __name: str) -> any:
        if __name.endswith('__context'):
            return self.__dict__['__context']
        if __name in self.__context:
            return self.__context[__name].get(None)
        raise AttributeError(f'attribute {__name} does not exist')

    def __setattr__(self, __name: str, __value: any) -> None:
        if __name.endswith('__context'):
            self.__dict__['__context'] = __value
        elif __name in self.__context:
            self.__context[__name].set(__value)
        else:
            raise AttributeError(f'attribute {__name} does not exist')


class Context(_ContextContainer):
    """
    This context aware class represents the "application context".
    The intention is to stuff values in this class with "inject"
    that is scoped internally to different things. Think of it
    as a dependency injection / inversion of code utility class.
    """
    def __init__(self) -> None:
        super().__init__('application', [
            'internal_scopes',
        ])

    def inject(self, name, value, scope='GLOBAL', force=False):
        """
        Inject a value, by name, in a specific scope (defaulting to GLOBAL).
        If a name is already in use a warning is emitted. Use the force
        flag to overwrite the value. This is useful in testing to override
        values intended for a live system in unit or integration tests.

        app_context.inject('myname', 'myvalue')
        assert app_context.resolve('GLOBAL')['myname'] == 'myvalue'
        """
        if self.internal_scopes is None:
            self.internal_scopes = {}
        if scope not in self.internal_scopes:
            self.internal_scopes[scope] = {}
        if force or name not in self.internal_scopes[scope]:
            self.internal_scopes[scope][name] = value
        else:
            logger.warning(f'Trying to override {name}, use remove or force')

    def remove(self, name, scope='GLOBAL'):
        """
        Explicitly removes a value injected in a scope.

        app_context.remove('myname')
        """
        return self.internal_scopes[scope].pop(name)

    def scopes(self):
        """
        Lists the available scopes the Context is tracking.

        app_context.scopes() == ['GLOBAL']
        """
        if self.internal_scopes is None:
            return []
        return self.internal_scopes.keys()

    def keys_in_scope(self, scope="GLOBAL"):
        """
        Lists the names within a scope the Context is tracking.

        app_context.keys_in_scope() == ['myname']
        """
        return self.resolve(scope).keys()

    def resolve(self, scope="GLOBAL"):
        """
        Returns the cache of values injected in a specified scope.

        app_context.resolve() == {'myname': 'myvalue'}
        """
        if self.internal_scopes is None:
            return {}
        if scope not in self.internal_scopes:
            return {}
        return self.internal_scopes[scope]


class Request(_ContextContainer):
    """
    This context aware class is intended to map to an incoming
    AWS Lambda invocation. The input "event" and "context" are
    set by the router, as well as cookie, query parameters, and
    header information.
    """
    def __init__(self) -> None:
        super().__init__("request", [
            "headers",
            "cookies",
            "body",
            "queryparams",
            "event",
            "context"
        ])

    def request_context(self, key, default_value=None):
        """
        This method is a helper to pull a key out of the
        "requestContext", which is AWS LAmbda hydrated metdata
        frequently used in websocket and HTTP API endpoints.
        """
        return self.event['requestContext'].get(key, default_value)

    def authorizer(self):
        return self.request_context('authorizer')

    def username(self):
        return self.authorizer()['jwt']['claims']['username']

    def account_id(self, osenv_hint='ACCOUNT_ID'):
        return self.request_context('accountId', os.getenv(osenv_hint))

    def api_id(self):
        return self.request_context('apiId')

    def remote_addr(self):
        return self.request_context('http')['sourceIp']

    def method(self):
        return self.request_context('http')['method']


class Response(_ContextContainer):
    """
    This context aware class represents the response data used
    in Router decorated methods. This includes the status_code,
    headers, and any body information.
    """
    def __init__(self) -> None:
        super().__init__("response", [
            "abort",
            "headers",
            "body",
            "status_code"
        ])

    def is_aborted(self):
        """
        Helper method to determine if the response was aborted
        in mid execution.
        """
        return self.abort is True

    def break_continuation(self):
        """
        This method is intended to be used by filter decorations
        from a Router instance to terminate execution and immediately
        dispatch output. This allows an application to stack multiple
        filters and conditionally invoke break_continuation() from
        further execution.
        """
        self.abort = True
        return self



from base64 import b64decode, b64encode
from decimal import Decimal
from Crypto.Cipher import AES
from Crypto.Hash import SHA256
import json


class TokenMarshaller:
    def encrypt(self, hash_key, header, last_key):
        pass

    def decrypt(self, hash_key, header, next_token):
        pass


class EncryptedTokenMarshaller(TokenMarshaller):
    def __init__(self, mode=AES.MODE_GCM) -> None:
        self.__mode = mode

    def __generate_key(self, account_id):
        return SHA256.new(bytes(account_id, 'utf-8')).digest()

    def __to_dto_key(self, last_key):
        new_val = {}
        for key, value in last_key.items():
            val = value
            if isinstance(val, Decimal):
                val = {'__internal': 'Decimal', 'value': int(value)}
            new_val[key] = val
        return new_val

    def __from_dto_key(self, last_key):
        new_val = {}
        for key, value in last_key.items():
            val = value
            if isinstance(val, dict) and '__internal' in val:
                val = globals()[val['__internal']](val['value'])
            new_val[key] = val
        return new_val

    def encrypt(self, hash_key, header, last_key):
        if last_key is None:
            return None
        cipher = AES.new(
            key=self.__generate_key(hash_key),
            mode=self.__mode)
        cipher.update(bytes(header, 'utf-8'))
        data = bytes(json.dumps(self.__to_dto_key(last_key)), 'utf-8')
        ciphertext, tag = cipher.encrypt_and_digest(data)
        token_payload = {
            'nonce': b64encode(cipher.nonce).decode('utf-8'),
            'payload': b64encode(ciphertext).decode('utf-8'),
            'tag': b64encode(tag).decode('utf-8')
        }
        token_bytes = bytes(json.dumps(token_payload), 'utf-8')
        return b64encode(token_bytes).decode('utf-8')

    def decrypt(self, hash_key, header, next_token):
        if next_token is None:
            return None
        values = json.loads(b64decode(next_token).decode('utf-8'))
        cipher = AES.new(
            key=self.__generate_key(hash_key),
            mode=self.__mode,
            nonce=b64decode(values['nonce']))
        cipher.update(bytes(header, 'utf-8'))
        plaintext = cipher.decrypt_and_verify(
            b64decode(values['payload']),
            b64decode(values['tag']))
        return self.__from_dto_key(json.loads(plaintext))



from ophis.database import Repository, QueryParams
from pytest import fail


def test_crud(thing_data):
    # Nothing here
    assert thing_data.get('012345678912', item_id="test") is None
    # Create
    things = []
    for index in range(0, 9):
        thing = thing_data.create('012345678912', item={
            'thingName': f'MyThing{index}',
        })
        assert thing['thingName'] == f'MyThing{index}'
        things.append(thing)
    try:
        thing_data.create('012345678912', item={
            'thingName': 'MyThing0'
        })
        fail('Should not have gotten here')
    except Exception as e:
        assert str(e) == 'The Things item already exists.'
    # Read
    assert things[0] == thing_data.get('012345678912', item_id='MyThing0')
    # List
    assert things == thing_data.items('012345678912').items
    # Pagination
    resp = thing_data.items('012345678912', params=QueryParams(limit=3))
    assert things[0:3] == resp.items
    assert resp.next_token is not None
    # Next Page
    resp = thing_data.items('012345678912', params=QueryParams(limit=3, next_token=resp.next_token))
    assert things[3:6] == resp.items
    assert resp.next_token is not None
    # Delete
    thing_data.delete('012345678912', item_id="MyThing0")
    assert thing_data.get('012345678912', item_id='MyThing0') is None
    try:
        thing_data.update('012345678912', item={'thingName': 'MyThing0'})
        fail('Should not have gotten here')
    except Exception as e:
        assert str(e) == "The Things item does not exist."
    # Batch Read
    batched = Repository.batch_read('012345678912', reads=[
        {
            'id': 'Nothing',
        },
        {
            'repository': thing_data,
            'id': 'MyThing1',
        }
    ])
    assert things[1:2] == batched
    # Batch Write
    Repository.batch_write('012345678912', updates=[
        {
            'id': 'Nothing',
        },
        {
            'repository': thing_data,
            'item': {
                'thingName': 'MyBatchedThing0'
            }
        },
        {
            'repository': thing_data,
            'item': {
                'thingName': 'MyThing1'
            },
            'delete': True
        }
    ])
    assert thing_data.get('012345678912', item_id='MyBatchedThing0') is not None
    assert thing_data.get('012345678912', item_id='MyThing1') is None



from decimal import Decimal
from ophis.token import EncryptedTokenMarshaller

TEST_ACC = "012345678912"


def test_roundtrip_failure():
    marshaller = EncryptedTokenMarshaller()
    token = marshaller.encrypt(TEST_ACC, header="This.Test", last_key={
        "message": 'This is huge!',
        "createTime": Decimal(100)
    })
    try:
        marshaller.decrypt("987654321021", "This.Test", token)
        raise Exception("Not supposed to!")
    except ValueError as e:
        assert "MAC check failed" == str(e)


def test_roundtrip():
    marshaller = EncryptedTokenMarshaller()
    data = {'message': 'This is huge!', 'createTime': Decimal(100)}
    token = marshaller.encrypt(TEST_ACC, "This.Test", data)
    assert data == marshaller.decrypt(TEST_ACC, "This.Test", token)



from contextvars import copy_context
from decimal import Decimal
import traceback
from ophis.globals import request, response, app_context
import inspect
import json
import logging
import re


logger = logging.getLogger(__name__)


class RouterEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            return f'{o.normalize():f}'
        return super().default(o)


class Router:
    """
    The router class used for invoking contextual aware invocations
    of API Gateway routes in the form of Python functions. The public
    methods on a Router instance are used to decorate the methods
    accordingly. The structure of the application and routes are
    up to the user to define.
    """
    def __init__(self, encoder=RouterEncoder) -> None:
        self.routes = {}
        self.filters = []
        self.encoder = encoder

    def __call__(self, event, context):
        """
        This magic method is the entry point to Lambda handlers typically used
        as "Singleton" functions to handle multiple routes. This means the
        "handler" parameter to an AWS Lambda function references the instance
        of a Router class used as decorations.
        """
        logger.debug(f'Incoming event {event}')
        ctx = copy_context()
        self.__prepare_context(ctx, event, context)
        for filter in self.filters:
            kwargs = self.__fill_globals(filter)
            if hasattr(filter, 'routeKey'):
                output = self.__dispatch_route_key(ctx, filter, **kwargs)
            else:
                output = ctx.run(filter, **kwargs)
            if self.__is_aborted(ctx):
                return self.__dispatch_response(ctx, output, aborted=True)
        for rule, route in self.routes.items():
            method, pattern = rule.split(':', 1)
            if method == event['requestContext']['http']['method']:
                raw_path = event['requestContext']['http']['path']
                match = re.search(pattern, raw_path)
                if pattern == raw_path or match is not None:
                    logger.info(f'Found {route.__module__}.{route.__name__}')
                    output = self.__dispatch_request(
                        ctx, route, list(match.groups()))
                    return self.__dispatch_response(ctx, output)
        return {
            'statusCode': 404,
            'headers': {'content-type': 'application/json'},
            'body': '{"message": "Resource not found"}'
        }

    def __dispatch_route_key(self, ctx, filter, **kwargs):
        route_key = getattr(filter, 'routeKey')

        def wrapper():
            if request.request_context('routeKey') == route_key:
                response.break_continuation()
                return filter(**kwargs)

        return ctx.run(wrapper)

    def __dispatch_response(self, ctx, output, aborted=False):
        def format_output(real_out):
            if isinstance(real_out, dict):
                real_out = json.dumps(real_out, cls=self.encoder)
                if 'content-type' not in response.headers:
                    response.headers['content-type'] = 'application/json'
            elif isinstance(real_out, str):
                if 'content-type' not in response.headers:
                    response.headers['content-type'] = 'text/plain'
            b = real_out if real_out is not None else response.body
            response.headers['content-length'] = 0 if b is None else len(b)
            if not aborted and b is None and response.status_code == 200:
                response.status_code = 204
            return {
                'statusCode': response.status_code,
                'body': b,
                'headers': response.headers
            }
        return ctx.run(format_output, output)

    def __fill_globals(self, route):
        kwargs = {}
        sig = inspect.signature(route)
        for key, value in app_context.resolve('GLOBAL').items():
            if key in sig.parameters:
                kwargs[key] = value
        return kwargs

    def __is_aborted(self, ctx):
        def obtain_aborted():
            return response.is_aborted()
        return ctx.run(obtain_aborted)

    def __fail(self, ctx):
        def fail_response():
            response.status_code = 500
        ctx.run(fail_response)

    def __dispatch_request(self, ctx, route, path_values):
        kwargs = self.__fill_globals(route)
        for index in range(0, len(route.path_names)):
            kwargs[route.path_names[index]] = path_values[index]
        try:
            return ctx.run(route, **kwargs)
        except Exception as e:
            trace = ''.join(traceback.format_exception(
                e,
                value=e,
                tb=e.__traceback__))
            logger.error(f"Failed to run {route.__name__}:\n{trace}")
            self.__fail(ctx)
            return {
                'message': 'Internal server error'
            }

    def __prepare_context(self, ctx, event, context):
        def fill_values():
            qparams = 'queryStringParameters'
            response.status_code = 200
            response.headers = {}
            response.abort = False
            request.cookies = event.get('cookies', {})
            request.headers = event.get('headers', {})
            request.queryparams = event.get(qparams, {})
            request.body = event['body'] if 'body' in event else None
            request.event = event
            request.context = context
        ctx.run(fill_values)

    def filter(self):
        """
        A filter gets invoked on every request handled by this router.
        Filter methods allow for terminations via the reponse global
        and the break_continuation() method. This is useful for
        appropriately handling the CORS headers and terminating an OK
        status. Example:

        @api.filter()
        def cors():
            response.headers = {
                'access-control-allow-origin': '*',
                'access-control-allow-methods': 'GET, POST, PUT, DELETE',
                'access-control-allow-headers': ', '.join([
                    'Content-Type',
                    'Content-Length',
                    'Authorization'
                ])
            }
            if request.method() == 'OPTIONS':
                response.break_continuation()
        """
        def wrapper(func):
            self.filters.append(func)
            return func
        return wrapper

    def routeKey(self, routeKey):
        """
        The routeKey method adds a filter function that terminations
        on a request routeKey. This is useful for websocket based integrations
        that target special actions like $connect and $disconnect:

        @api.routeKey('$connect')
        def connect(connection_data):
            connection_data.create(
                request.account_id(),
                item={
                    'connectionId': request.request_context('connectionId')
                })

        @api.routeKey('$disconnect')
        def disconnect(connection_data):
            connection_data.delete(
                request.account_id(),
                item_id=request.request_context('connectionId'),
            )
        """
        def wrapper(func):
            setattr(func, 'routeKey', routeKey)
            self.filters.append(func)
            return func
        return wrapper

    def route(self, path, methods=["GET"]):
        """
        The decoration allows functions to target traditional HTTP methods.
        The path can be parameterized, and injected as inputs to the
        decorated method. Example:

        @api.route('/greet/{name}')
        def greet(name):
            return {
                'message': f'Hello, {name}'
            }
        """
        def wrapper(func):
            for method in methods:
                path_parts = []

                def replace_update(obj):
                    path_parts.append(obj.group(0).lstrip(':'))
                    return '([^/]+)'
                new_path = '^' + re.sub(':[^/]+', replace_update, path) + '$'
                rule = ':'.join([method.upper(), new_path])
                if rule in self.routes:
                    logger.warning(f'Rule {rule} is already in use.')
                setattr(func, 'path_names', path_parts)
                self.routes[rule] = func
            return func
        return wrapper



#!/bin/env python3

from setuptools import setup, find_packages


with open('README.md', 'r') as f:
    long_description = f.read()


setup(
    name="ophis-py",
    version="0.1.0",
    description="Simple Python bindings for AWS applications",
    long_description=long_description,
    long_description_context_type="text/markdown",
    author="Philip Cali",
    author_email="philip.cali@gmail.com",
    url="https://github.com/philcali/ophis-py",
    license="Apache License 2.0",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent"
    ],
    install_requires=[
        'boto3',
        'pycryptodome'
    ],
    extra_require={
        'test': ['pytest']
    }
)



from ophis.router import Router
from ophis.globals import request, response, app_context
import json
import os


api = Router()
app_context.inject("some_key", "some_value")


@api.filter()
def cors():
    if request.method() == 'OPTIONS':
        response.headers = {
            'access-control-allow-origin': '*',
            'access-control-allow-methods': 'GET, POST, PUT, DELETE',
            'access-control-allow-headers': ', '.join([
                'Content-Type',
                'Content-Length',
                'Authorization'
            ])
        }
        response.break_continuation()


@api.route("/")
def index(some_key):
    assert some_key == "some_value"
    return "Hello World"


@api.routeKey('$connect')
def connect():
    return "Hello World"


@api.route("/other/path")
def other_path(some_key):
    assert some_key == "some_value"
    return "Hello World"


@api.route('/pets/:petId')
def get_pet(petId):
    return f'Hello World, {petId}'


@api.route('/pets/:petId', methods=['PUT', 'POST'])
def put_pet(petId):
    assert 'value1' == request.authorizer()['jwt']['claims']['claim1']
    pet = json.loads(request.body)
    assert pet['name'] == 'Pixie'
    return {
        'id': petId,
        'name': pet['name']
    }


@api.route('/pets/:petId', methods=['DELETE'])
def delete_pet(petId):
    assert petId == 'pixie'
    response.status_code = 204


def test_router():
    events = [file for file in os.listdir('events') if ".json" in file]
    for test_event in events:
        with open(os.path.join('events', test_event)) as f:
            event = json.loads(f.read())
        assert event['response'] == api(event=event['event'], context={})



from decimal import Decimal
from math import floor
from time import time
from collections import namedtuple
from boto3.dynamodb.conditions import Key, And
from botocore.exceptions import ClientError
from ophis.globals import app_context
from ophis.token import EncryptedTokenMarshaller
import logging


MAX_ITEMS = 100
CON_CHECK_CODE = 'ConditionalCheckFailedException'

logger = logging.getLogger(__name__)

SortFilter = namedtuple(
    'SortFilter',
    field_names=['name', 'method', 'values'])
QueryParams = namedtuple(
    'QueryParams',
    field_names=['limit', 'next_token', 'sort_filters', 'sort_ascending'],
    defaults=[MAX_ITEMS, None, [], True])
QueryResults = namedtuple(
    'QueryResults',
    field_names=['items', 'next_token'],
    defaults=[[], None])


class ConflictException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class NotFoundException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class Repository():
    def __init__(
            self,
            type,
            table=None,
            fields_to_keys={},
            token_marshaller=EncryptedTokenMarshaller()) -> None:
        resolved = app_context.resolve('GLOBAL')
        self.table = table if table is not None else resolved['table']
        self.type = type
        self.fields_to_keys = fields_to_keys
        self.tokens = token_marshaller

    def batch_read(*args, reads, ddb=None, table=None):
        if table is None:
            table = app_context.resolve('GLOBAL')['table']
        if ddb is None:
            ddb = app_context.resolve('GLOBAL')['dynamodb']
        params = {table.name: {'Keys': []}}
        lookups = {}
        for entry in reads:
            if 'repository' not in entry or 'id' not in entry:
                logger.warning(f'Read skipped to missing fields: {entry}')
                continue
            keys = list(args)
            if 'parent_ids' in entry:
                for parent_id in entry.pop('parent_ids'):
                    keys.append(parent_id)
            key = {
                'PK': entry['repository'].make_hash_key(*keys),
                'SK': entry['id']
            }
            params[table.name]['Keys'].append(key)
            lookups[key['PK'] + ':' + key['SK']] = entry['repository']
        resp = ddb.batch_get_item(RequestItems=params)
        return [
            lookups[item['PK'] + ':' + item['SK']].prune_dto(item)
            for item in resp['Responses'][table.name]
        ]

    def batch_write(*args, updates, table=None):
        if table is None:
            table = app_context.resolve('GLOBAL')['table']
        with table.batch_writer() as batch:
            for update in updates:
                if 'repository' not in update or 'item' not in update:
                    logger.warning(f'Update skipped to missing fields: {update}')
                    continue
                keys = list(args)
                if 'parent_ids' in update:
                    for parent_id in update.pop('parent_ids'):
                        keys.append(parent_id)
                dto = update['repository'].make_dto(
                    *keys,
                    item=update['item'],
                    time_fields=['create', 'update'])
                if update.get("delete", False):
                    batch.delete_item(Key={'PK': dto['PK'], 'SK': dto['SK']})
                else:
                    batch.put_item(Item=dto)

    def prune_dto(self, original):
        if original is None:
            return None
        rval = {}
        for key, value in original.items():
            if key in ['PK', 'SK', 'GS1-PK']:
                continue
            if type(value) is Decimal:
                value = int(value)
            rval[key] = value
        return rval

    def make_dto(self, *args, item, time_fields=['create', 'update']):
        new_item = {
            'PK': self.make_hash_key(*args)
        }
        for key, value in item.items():
            new_item[key] = value
        for key, value in self.fields_to_keys.items():
            if key not in item:
                raise Exception(f'The {self.type} item needs a {key} field')
            new_item[value] = item[key]
        for time_field in time_fields:
            if f'{time_field}Time' not in item:
                new_item[f'{time_field}Time'] = int(floor(time()))
        return new_item

    def __encrypt_token(self, hash_key, res):
        return self.tokens.encrypt(
            hash_key=hash_key,
            header=':'.join([self.type, 'next_token']),
            last_key=res.get('LastEvaluatedKey', None))

    def make_hash_key(self, *args):
        return ":".join([self.type] + list(args))

    def __list(self, *args, **kwargs):
        hash_key = self.make_hash_key(*args)
        normal_key = 'PK'
        q_params = {'ScanIndexForward': kwargs['params'].sort_ascending}
        if 'index_name' in kwargs:
            q_params['IndexName'] = kwargs['index_name']
            normal_key = f'{kwargs["index_name"]}-PK'
        key_cond = Key(normal_key).eq(hash_key)
        for f in kwargs['params'].sort_filters:
            sk = f.name
            if f.name in self.fields_to_keys:
                sk = self.fields_to_keys[f.name]
            key_cond = And(key_cond, getattr(Key(sk), f.method)(*f.values))
        q_params['KeyConditionExpression'] = key_cond
        q_params['Limit'] = kwargs['params'].limit
        last_key = self.tokens.decrypt(
            hash_key=hash_key,
            header=':'.join([self.type, 'next_token']),
            next_token=kwargs['params'].next_token)
        if last_key is not None:
            q_params['ExclusiveStartKey'] = last_key
        response = self.table.query(**q_params)
        items = map(
            self.prune_dto,
            response['Items'] if 'Items' in response else [])
        return QueryResults(
            items=list(items),
            next_token=self.__encrypt_token(hash_key, response)
        )

    def items(self, *args, params=QueryParams()):
        kwargs = {'params': params}
        return self.__list(*args, **kwargs)

    def items_index(self, *args, index_name, params=QueryParams()):
        kwargs = {'params': params, 'index_name': index_name}
        return self.__list(*args, **kwargs)

    def create(self, *args, item):
        new_item = self.make_dto(*args, item=item)
        exp = "attribute_not_exists(PK) and attribute_not_exists(SK)"
        try:
            self.table.put_item(
                Item=new_item,
                ConditionExpression=exp
            )
        except ClientError as e:
            if e.response['Error']['Code'] == CON_CHECK_CODE:
                raise ConflictException(
                    f'The {self.type} item already exists.')
            raise e
        return self.prune_dto(new_item)

    def update(self, *args, item, fields_to_remove=[]):
        new_item = self.make_dto(*args, item=item, time_fields=['update'])
        condition = "attribute_exists(PK) and attribute_exists(SK)"
        update_exp = []
        remove_exp = []
        update_names = {}
        update_values = {}
        for key, value in new_item.items():
            if key in ['PK', 'SK'] or key in self.fields_to_keys:
                continue
            update_names[f'#{key}'] = key
            update_values[f':{key}'] = value
            update_exp.append(f'#{key} = :{key}')
        for field in fields_to_remove:
            update_names[f'#{field}'] = field
            remove_exp.append(f'#{field}')
        try:
            expression = []
            if len(update_exp) > 0:
                expression.append(f'SET {", ".join(update_exp)}')
            if len(remove_exp) > 0:
                expression.append(f'REMOVE {", ".join(remove_exp)}')
            response = self.table.update_item(
                Key={
                    'PK': new_item['PK'],
                    'SK': new_item['SK']
                },
                ConditionExpression=condition,
                UpdateExpression=" AND ".join(expression),
                ExpressionAttributeNames=update_names,
                ExpressionAttributeValues=update_values,
                ReturnValues='ALL_NEW',
            )
            return self.prune_dto(response['Attributes'])
        except ClientError as e:
            if e.response['Error']['Code'] == CON_CHECK_CODE:
                raise NotFoundException(
                    f'The {self.type} item does not exist.')
            raise e

    def delete(self, *args, item_id):
        self.table.delete_item(
            Key={
                'PK': self.make_hash_key(*args),
                'SK': item_id
            }
        )

    def get(self, *args, item_id):
        response = self.table.get_item(
            Key={
                'PK': self.make_hash_key(*args),
                'SK': item_id
            }
        )
        return self.prune_dto(response.get('Item', None))
